# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

import pytest

from pyrubrum import DictDatabase
from pyrubrum import NotFoundError

database = DictDatabase()


def test_set():
    database.set("A", "B")

    assert "A" in database
    assert database["A"] == "B"


def test_get():
    assert database.get("A") == "B"

    with pytest.raises(NotFoundError):
        database.get("C")


def test_delete():
    database.delete("A")
    assert "A" not in database

    with pytest.raises(NotFoundError):
        database.delete("A")



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from typing import Set

from environs import Env
from pyrogram import Client

from pyrubrum import Handler, Menu, Node, transform

tree = transform(
    {
        Menu("Start", "start", "Hello!", default=True): [
            Menu("About me", "about_me", "I'm just a bot!"),
            Menu(
                "Thoughts", "thoughts", "I'm a bot, I cannot think properly..."
            ),
        ]
    }
)


def main(
    api_hash: str,
    api_id: int,
    bot_token: str,
    session_name: str,
    tree: Set[Node],
):
    bot = Client(
        session_name, api_hash=api_hash, api_id=api_id, bot_token=bot_token
    )
    handler = Handler(tree)
    handler.setup(bot)

    bot.run()


if __name__ == "__main__":
    env = Env()
    env.read_env()

    api_hash = env("API_HASH")
    api_id = env.int("API_ID")
    bot_token = env("BOT_TOKEN")
    session_name = env("SESSION_NAME")

    main(api_hash, api_id, bot_token, session_name, tree)



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.


class DatabaseError(Exception):
    """General exception which catches all the errors related to database
    issues.
    """

    pass



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Any, Dict, List

from pyrogram import Client


class BaseStyle(ABC):
    """Basic representation of a style that can be used in order to
    create a keyboard for `Menu.keyboard`.
    """

    @abstractmethod
    def generate(
        self,
        handler: "Handler",  # noqa
        client: Client,
        context: Any,
        parameters: Dict[str, Any],
        menu: "Menu",  # noqa
    ) -> List[List["Button"]]:
        """This abstract method is intended to be implemented as a keyboard
        generator.

        Parameters:
            handler (BaseHandler): The handler which coordinates the management
                of the menus.
            client (Client): The client which is linked to the handler.
            context (Union[CallbackQuery, Message]): The context for which the
                button is generated.
            parameters (Dict[str, Any]): The parameters which were passed to
                the handler.
            menu (Menu): The menu the keyboard is being built for.

        Returns:
            pyrogram.InlineKeyboardMarkup: The generated inline keyboard,
            which is then displayed to the user.
        """
        raise NotImplementedError



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Union

from pyrogram.filters import Filter
from pyrogram.types import InputMedia

from pyrubrum.menus.styles import BaseStyle, PageStyle
from pyrubrum.types import Types

from .menu import Menu


class PageMenu(Menu):
    """Implementation of a menu which automatically, given a list of items, manages
    paging and the setting of parameters. It has got, by definition, all the
    parameters defined in `Menu` plus a list of items that will be displayed
    to the user.

    See Also:
        For understanding how the keyboard is being generated, see `PageStyle`.

    Parameters:
        name (str): The name you give to the menu, which will be used as
            the text of callback button, if needed. See `BaseMenu` for more
            information.
        menu_id (str): The unique identifier given to the menu, which will
            refer unequivocally to this entity. The hash for this class is
            generated relying on the content of this field. Avoid using ``0``
            as it is used for buttons whose purpose is only related to design
            (i.e. they do not point to any menu, see :term:`Null-pointer
            button`). See `BaseMenu` for more information.
        content (Types.Content): What will be displayed whenever a user
            accesses this menu. Both text and media can be provided. A
            function can be provided as well and must follow the following
            arguments pattern::

                func(handler, client, context, parameters)

            See `Menu` for more information.
        items (types.Items): The list of elements the menu is compounded of or
            a function which returns such type of value.
        deep_link (Optional[bool]): If this menu shall be reached by a
            deep-link whose payload is the identifier of this instance.
            Defaults to ``False``.
        default (Optional[bool]): If this menu shall be displayed if no
            other :term:`top-level menu <Top-level menu>` has been matched.
            It works only if this menu is a :term:`top-level <Top-level menu>`
            one.
        message_filter (Optional[Filter]): A filter for telling Pyrogram
            when a message should be associated to this menu. It works only
            for top-level menus (see `Handler.setup`). Defaults to ``None``,
            which automatically makes this menu reachable when the user texts
            a message that follows this pattern::

                /[MENU_ID]

        preliminary (Types.Preliminary): A function which is executed each time
            before doing anything else in `on_callback` and `on_message`.
            You can provide a list of such functions as well, which will be
            executed following the same order as the one of the list.
            Defaults to ``None``, which means that no function is going to
            be executed.
        style (BaseStyle): The class which generates the keyboard for this
            function following a certain style. Defaults to `PageStyle()`.

    Warning:
        This implementation is not compatible with a non-parameterized handler.
        An handler that supports parameterization is required.
    """

    def __init__(
        self,
        name: str,
        menu_id: str,
        content: Union[InputMedia, str],
        items: Types.Items,
        deep_link: Optional[bool] = False,
        default: Optional[bool] = False,
        message_filter: Optional[Filter] = None,
        preliminary: Types.Preliminary = None,
        style: BaseStyle = PageStyle(),
    ):
        """Roughly equivalent to::

            Menu(self, name, menu_id, content, default, message_filter,
                 preliminary, style, items=items)
        """
        Menu.__init__(
            self,
            name,
            menu_id,
            content,
            deep_link=deep_link,
            default=default,
            message_filter=message_filter,
            preliminary=preliminary,
            style=style,
        )

        self.items = items



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from pyrubrum.types import Types

from .base_database import BaseDatabase
from .errors import DeleteError, ExpireError, NotFoundError, SetError

try:
    import redis
except (ImportError, ModuleNotFoundError):
    pass


class RedisDatabase(BaseDatabase):
    """Implementation of a database using a Redis server.

    Parameters:
        server (redis.Redis): The Redis instance which is being used.
        encoding (Optional[str]): The encoding format which shall be used
            to decode the content that is retrieved from the Redis server.
            Defaults to ``utf-8``.
        default_expire (Types.Expire): The expire which is set by default. If
            it is ``False``, no expire shall be set. Defaults to ``86400``
            seconds (i.e. a day).
    """

    def __init__(
        self,
        server: "redis.Redis",
        encoding: Optional[str] = "utf-8",
        default_expire: Optional[Types.Expire] = 86400,
    ):
        self.default_expire = default_expire
        self.encoding = encoding
        self.server = server

    def get(self, key: str) -> str:
        """Get the value which is associated to a certain key inside the database.
        If no key is found, `NotFoundError` is raised.

        This method will query the key using `redis.Redis.get`.

        Parameters:
            key (str): The key you are retrieving the value of from the
                Redis database.

        Returns:
            Optional[str]: The value which is associated to the key in the
            database.

        Raises:
            NotFoundError: If the provided key is not found.
        """
        content = self.server.get(key)

        if not content:
            raise NotFoundError(key)

        return content.decode(self.encoding)

    def set(self, key: str, value: str, expire: Types.Expire = None):
        """Assign a value to a certain key inside the database. If no expire is
        provided, it will automatically provide one from `default_expire`. If
        the expire is set to be ``False``, no expire flag will be assigned.

        This method will assign the provided value to the key using
        `redis.Redis.set` and mark the expire flag with `redis.Redis.expire`.

        Parameters:
            key (str): The key you are adding or updating the value of.
            value (str): The value which is being assigned to the key.
            expire (Optional[Types.Expire]): The expire in seconds or as a
                `timedelta` object. A key is set not to expire if ``False`` is
                provided for this argument. Defaults to ``None``, which
                automatically provides a default expire from `default_expire`.

        Raises:
            ExpireError: If an error occured while setting the expire for the
                key.
            SetError: If an error occured while inserting the key into the
                database.
        """
        if not self.server.set(key, value):
            raise SetError

        if expire and not self.server.expire(key, expire):
            raise ExpireError
        elif (
            expire is not False
            and self.default_expire
            and not self.server.expire(key, self.default_expire)
        ):
            raise ExpireError

    def delete(self, key: str):
        """Delete a certain key from the database, together with its stored value.

        This method will delete the provided key from the database using
        `redis.Redis.delete`.

        Parameters:
            key (str): The key which is being deleted from the database,
                together with its linked data.

        Raises:
            DeleteError: If an error occured while deleting the key.
            NotFoundError: If the provided key is not found.
        """
        try:
            if not self.server.delete(key):
                raise NotFoundError(key)
        except redis.RedisError:
            raise DeleteError



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

import pytest

from pyrubrum import DeepLinkMenu
from pyrubrum.menus.deep_link_menu import PERMITTED_TYPES, USERNAMES

client_test = "Test client"
username_test = "test_bot"

USERNAMES[hash(client_test)] = username_test

payload = "test_payload"

expected_deep_link = "https://t.me/%s?start=%s" % (username_test, payload)


@pytest.mark.parametrize("deep_type", PERMITTED_TYPES | {"fancytype"})
def test_deep_link_menu(deep_type):
    if deep_type in PERMITTED_TYPES:
        menu = DeepLinkMenu("Test", "test", payload, deep_type)
        menu.button(None, client_test, None)
    else:
        with pytest.raises(ValueError):
            DeepLinkMenu("Test", "test", payload, deep_type)


def test_button():
    menu = DeepLinkMenu("Test", "test", payload)
    button = menu.button(None, client_test, None)

    assert button.link == expected_deep_link
    assert not hasattr(button, "menu_id")
    assert button.name == "Test"


def test_button_with_function():
    menu = DeepLinkMenu("Test", "test", lambda a, b, c, d: payload)
    button = menu.button(None, client_test, None)

    assert button.link == expected_deep_link
    assert not hasattr(button, "menu_id")
    assert button.name == "Test"



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from typing import List, Set, Union

from environs import Env
from pyrogram import Client

from pyrubrum import (
    DeepLinkMenu,
    DictDatabase,
    Element,
    Menu,
    Node,
    PageMenu,
    ParameterizedHandler,
    RedisDatabase,
    transform,
)

try:
    from redis import Redis
except (ImportError, ModuleNotFoundError):
    pass

drinks = [
    "Beer",
    "Cider",
    "Coffee",
    "Coke",
    "French wine",
    "Gin",
    "Hard soda",
    "Irish beer",
    "Italian wine",
    "Liquor",
    "Water",
    "Wine",
]
snacks = [
    "Bubblegum",
    "Candies",
    "Chocolate bar",
    "Crisps",
    "Cracker nuts",
    "Crêpe",
    "Pancake",
    "Pizza",
    "Pretzel",
    "Waffle",
]


def make_elements(elements: List[str]) -> List[Element]:
    return [Element(name, index) for index, name in enumerate(elements)]


tree = transform(
    {
        Menu(
            "Start",
            "start",
            "ℹ️ Have a drink or a snack by saying /drink or /snack",
            default=True,
        ): [
            DeepLinkMenu("🍷 Drink", "link_drink", "drink"),
            DeepLinkMenu("🍬 Snack", "link_snack", "snack"),
        ],
        PageMenu(
            "Drink",
            "drink",
            "🍷 Choose the drink you want!",
            make_elements(drinks),
            deep_link=True,
        ): {Menu("After drink choice", "drink_chosen", "🍹 Here you go!")},
        PageMenu(
            "Snack",
            "snack",
            "🍬 Choose the snack you want!",
            make_elements(snacks),
            deep_link=True,
        ): {Menu("After snack choice", "snack_chosen", "🥨 Here you go!")},
    }
)


def main(
    api_hash: str,
    api_id: int,
    bot_token: str,
    database: Union[DictDatabase, RedisDatabase],
    session_name: str,
    tree: Set[Node],
):
    bot = Client(
        session_name, api_hash=api_hash, api_id=api_id, bot_token=bot_token
    )

    handler = ParameterizedHandler(tree, database)
    handler.setup(bot)

    bot.run()


if __name__ == "__main__":
    env = Env()
    env.read_env()

    api_hash = env("API_HASH")
    api_id = env.int("API_ID")
    bot_token = env("BOT_TOKEN")
    session_name = env("SESSION_NAME")

    if env.bool("USE_REDIS", False):
        db = env.int("REDIS_DB")
        host = env("REDIS_HOST")
        port = env.int("REDIS_PORT")
        database = RedisDatabase(Redis(host=host, port=port, db=db))
    else:
        database = DictDatabase()

    main(api_hash, api_id, bot_token, database, session_name, tree)



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from itertools import islice
from math import ceil
from typing import Any, Dict, List, Optional

from pyrogram import Client

from pyrubrum.handlers.base_handler import NULL_POINTER
from pyrubrum.keyboard.button import Button

from .base_style import BaseStyle

EMPTY_CHARACTER = "‎"


class PageStyle(BaseStyle):
    """Integrate paging into your menu in a simple way by using this
    class.

    Parameters:
        back_text (Optional[str]): The text which will be displayed
            inside the button that lets the user go back to the parent
            menu. Defaults to "🔙".
        limit (Optional[int]): The limit of buttons per row. Defaults to ``2``.
        limit_items (Optional[int]): The limit of elements per page.
            Defaults to ``4``.
        next_page_text (Optional[str]): The text which is displayed
            inside the button that lets the user move on to the next page,
            if any. Defaults to "▶️".
        previous_page_text (Optional[str]): The text which is
            displayed inside the button that lets the user go back to the
            previous page, if any. Defaults to "◀️".
        show_counter (Optional[bool]): If there shall be the number of the
            total pages right next to the number of page. Defaults to
            ``True``.
        show_page (Optional[bool]): If the number of the page shall be
            displayed between the two arrows for managing the motion of
            the pages. Defaults to ``True``.

    Warning:
        This implementation is not compatible with a non-parameterized handler.
        An handler that supports parameterization is required.
    """

    def __init__(
        self,
        back_text: Optional[str] = "🔙",
        limit: Optional[int] = 2,
        limit_items: Optional[int] = 4,
        next_page_text: Optional[str] = "▶️",
        previous_page_text: Optional[str] = "◀️",
        show_counter: Optional[bool] = True,
        show_page: Optional[bool] = True,
    ):
        self.back_text = back_text
        self.limit = limit
        self.limit_items = limit_items
        self.next_page_text = next_page_text
        self.previous_page_text = previous_page_text
        self.show_counter = show_counter
        self.show_page = show_page

    def generate_page_id(
        self,
        handler: "ParameterizedHandler",  # noqa
        client: Client,
        context: Any,
        parameters: Dict[str, Any],
        menu: "PageMenu",  # noqa
    ) -> str:
        """Generate a page identifier, relying on the identifier of the provided menu,
        built in the following way::

            page_[MENU_ID]

        Hence, there is a special set of keys for parameters that is
        recommended not to be used in order to handle paging in a proper way.
        All keys starting with ``page_`` shall then be avoided.

        Parameters:
            handler (ParameterizedHandler): The handler which coordinates the
                management of the menus and supports parameterization.
            client (Client): The client which is linked to the handler.
            context (Union[CallbackQuery, Message]): The context for which the
                button is generated.
            parameters (Dict[str, Any]): The parameters which were passed to
                the handler.
            menu (Menu): The menu the keyboard is being built for.

        Returns:
            str: The generated page identifier.
        """
        page_id = "page_" + menu.menu_id

        if page_id not in parameters:
            parameters[page_id] = 0

        if "same_menu" in parameters and parameters["same_menu"]:
            parameters[page_id] = int(parameters["element_id"])

        return page_id

    def generate(
        self,
        handler: "ParameterizedHandler",  # noqa
        client: Client,
        context: Any,
        parameters: Dict[str, Any],
        menu: "Menu",  # noqa
    ) -> List[List["Button"]]:
        """Provide a keyboard which is created relying on the page provided by the
        parameters and is built following the limits defined during the
        initialization of this instance (i.e. `limit` and `limit_items`).

        The page identifier is retrieved from `PageStyle.generate_page_id`.

        If this menu has been referenced by this menu itself (i.e. the user
        clicked on the buttons for moving between pages), the page is set to be
        equal to ``element_id``, which represents the page itself.

        As a result of this, during the build of the the teleport buttons (i.e.
        the buttons for moving between pages) ``element_id`` is set to be
        equal to the page the buttons refer to (e.g. page plus one for "Go to
        the next page").

        No teleport keyboard is generated if the number of total pages is equal
        to one. Otherwise, three buttons are generated. The first one is built
        in order to let the user go back to the previous page, if any, else a
        :term:`null-pointer button <Null-pointer button>` is created, with its
        content being an empty character. The second one is generated if and
        only if `PageStyle.show_menu` is set to be ``True``, with its text
        being the current number of page (concatenated to the number of total
        pages if `PageStyle.show_counter` is set to be ``True``). The third
        one is designed for moving on to the next page, if any, and it is built
        following the same rules that were respected by the first button.

        In the end, the keyboard is filled with all the buttons which refer to
        the children menus a special button for linking the user to the parent
        menu, if any.

        Parameters:
            handler (ParameterizedHandler): The handler which coordinates the
                management of the menus and supports parameterization.
            client (Client): The client which is linked to the handler.
            context (Union[CallbackQuery, Message]): The context for which the
                button is generated.
            parameters (Dict[str, Any]): The parameters which were passed to
                the handler.
            menu (Menu): The menu the keyboard is being built for.

        Returns:
            List[List[Button]]: The generated inline keyboard, which is then
            displayed to the user.
        """
        page_id = self.generate_page_id(
            handler, client, context, parameters, menu
        )

        parent, children = handler.get_family(menu.menu_id)

        keyboard = []

        if children:
            iterable = iter(children)
            page_item_menu = next(iterable)

            if callable(menu.items):
                items = menu.items(handler, client, context, parameters)
            elif isinstance(menu.items, list):
                items = menu.items
            else:
                raise TypeError("items must be either callable or a list")

            elements = items[parameters[page_id] * self.limit_items :][
                : self.limit_items
            ]

            keyboard = [
                [
                    Button(
                        element.name,
                        page_item_menu.menu_id,
                        parameters,
                        element.element_id,
                    )
                    for element in elements[i : i + self.limit]
                ]
                for i in range(0, len(elements), self.limit)
            ]

        if children:
            keyboard += list(
                iter(
                    lambda: list(
                        map(
                            lambda child: child.button(
                                handler, client, context, parameters
                            ),
                            islice(iterable, self.limit),
                        )
                    ),
                    [],
                )
            )

        pages = ceil(len(items) / self.limit_items)

        if pages > 1:
            teleport_row = []

            if parameters[page_id] > 0:
                previous_page_button = Button(
                    self.previous_page_text,
                    menu.menu_id,
                    parameters,
                    element_id=parameters[page_id] - 1,
                    same_menu=True,
                )

                if menu.menu_id + "_id" in parameters:
                    previous_page_button.parameters[
                        menu.menu_id + "_id"
                    ] = parameters[menu.menu_id + "_id"]

                teleport_row.append(previous_page_button)
            else:
                previous_page_button = Button(
                    EMPTY_CHARACTER, str(parameters[page_id]), NULL_POINTER,
                )

                teleport_row.append(previous_page_button)

            if self.show_page:
                if self.show_counter:
                    page_button = Button(
                        "%d/%d" % (parameters[page_id] + 1, pages),
                        str(parameters[page_id]),
                        NULL_POINTER,
                    )

                    teleport_row.append(page_button)
                else:
                    page_button = Button(
                        str(parameters[page_id] + 1), NULL_POINTER
                    )

                    teleport_row.append(page_button)

            if parameters[page_id] + 1 < pages:
                next_page_button = Button(
                    self.next_page_text,
                    menu.menu_id,
                    parameters,
                    element_id=parameters[page_id] + 1,
                    same_menu=True,
                )

                if menu.menu_id + "_id" in parameters:
                    next_page_button.parameters[
                        menu.menu_id + "_id"
                    ] = parameters[menu.menu_id + "_id"]

                teleport_row.append(next_page_button)
            else:
                next_page_button = Button(
                    EMPTY_CHARACTER, str(parameters[page_id]), NULL_POINTER,
                )

                teleport_row.append(next_page_button)

            keyboard += [teleport_row]

        if parent:
            parent_button = parent.button(handler, client, context, parameters)
            parent_button.name = self.back_text

            keyboard += [[parent_button]]

        return keyboard



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from hashlib import md5
from json import JSONDecodeError
from typing import List

from pyrogram import Client
from pyrogram.filters import Filter, create
from pyrogram.handlers import CallbackQueryHandler
from pyrogram.types import CallbackQuery, InlineKeyboardButton

from pyrubrum.database import BaseDatabase
from pyrubrum.database.errors import DatabaseError
from pyrubrum.keyboard import Button
from pyrubrum.types import Types

from .base_handler import NULL_POINTER, BaseHandler

try:
    import orjson as json  # noqa
except (ImportError, ModuleNotFoundError):
    import json


class ParameterizedBaseHandler(BaseHandler):
    """Basic implementation of an handler which has got, by definition, a database,
    with which it is able to perform
    :term:`parameterization <Parameterization>` (i.e. it supports parameters).

    The purpose of this class is to give a general interface for an handler
    which supports parameterization, as it doesn't implement anything except
    for a basic setup of the client (see `setup`) and a filter that allows the
    recognition of parameterized queries (see `filter`).

    Parameters:
        database (BaseDatabase): The storage for all the query parameters.
            It is used to pass parameters between menus.
    """

    def __init__(self, database: BaseDatabase):
        self.database = database

    def filter(self, menu_id: str) -> Filter:
        """Generate a function which is used by `CallbackQueryHandler` in order to
        filter callback queries relying on the content of the provided menu
        identifier.

        The content of the callback query is always a MD5 hash which behaves
        as the key for the parameters that are associated to the query the
        user made.

        If the identifier of the chat from which the query was sent does not
        match the one defined in the retrieved parameters, the callback is
        considered invalid.

        The filter returns ``True`` if the query is valid and matches
        ``menu_id``. Otherwise, it returns ``False``.

        Parameters:
            menu_id (str): The unique identifier of the menu that has to be
                matched.

        Returns:
            Filter: The function that gets called whenever a callback query is
            being handled.
        """

        def func(_, __, callback: CallbackQuery):
            if not hasattr(callback, "_result_data"):
                try:
                    callback._result_data = json.loads(
                        self.database.get(callback.data)
                    )

                    if (
                        callback._result_data["from_chat_id"]
                        == callback.message.chat.id
                    ):
                        self.database.delete(callback.data)
                except (DatabaseError, JSONDecodeError):
                    return False

            if (
                callback._result_data["from_chat_id"]
                != callback.message.chat.id
            ):
                return False

            if callback._result_data["menu_id"] == menu_id:
                callback.parameters = callback._result_data

                if (
                    not callback.parameters["same_menu"]
                    and callback.parameters["element_id"]
                ):
                    callback.parameters[menu_id + "_id"] = callback.parameters[
                        "element_id"
                    ]

                return True

            return False

        return create(func, "ParameterizedCallbackData")

    def process_keyboard(
        self,
        keyboard: List[List[Button]],
        callback_query_id: str,
        chat_id: int,
    ) -> List[List[InlineKeyboardButton]]:
        """Given a list which represents an inline keyboard and a
        unique identifier for the callback, generate a Pyrogram-compatible
        inline keyboard.

        It generates a MD5 hash key for each button by following this pattern::

            [CALLBACK_QUERY_ID][MENU_ID][ELEMENT_ID]

        After having generated a key, it sets it to be equal, inside the
        database, to the parameters of the button, which have been
        previously converted to JSON and include ``from_chat_id``, the
        identifier of the chat from which the query was sent, ``element_id``,
        ``menu_id`` and ``same_menu``.

        Parameters:
            keyboard (List[List[Button]]): The inline keyboard you want to
                process.
            callback_query_id (str): The unique identifier of the callback
                for which the keyboard is generated.
            chat_id (int): The identifier of the chat from which the query has
                been sent.

        Returns:
            List[List[pyrogram.InlineKeyboardButton]]: The generated keyboard
            in a Pyrogram-compatible type.
        """

        processed_keyboard = []

        for row in keyboard:
            processed_row = []
            processed_keyboard.append(processed_row)

            for button in row:
                if button.link:
                    processed_row.append(
                        InlineKeyboardButton(button.name, url=button.link)
                    )
                    continue

                if button.menu_id == NULL_POINTER:
                    processed_row.append(
                        InlineKeyboardButton(
                            button.name, callback_data=NULL_POINTER
                        )
                    )
                    continue

                key = md5(
                    (
                        str(callback_query_id)
                        + button.menu_id
                        + str(button.element_id)
                    ).encode("utf-8")
                ).hexdigest()

                content = (
                    button.parameters
                    if isinstance(button.parameters, dict)
                    else {}
                )

                content.update(
                    {
                        "from_chat_id": chat_id,
                        "element_id": button.element_id,
                        "menu_id": button.menu_id,
                        "same_menu": button.same_menu,
                    }
                )

                processed_row.append(
                    InlineKeyboardButton(button.name, callback_data=key)
                )

                self.database.set(key, json.dumps(content))

        return processed_keyboard

    def setup(self, client: Client):
        """Make all the defined menus reachable by the client by adding handlers that
        catch all their identifiers to it. It adds support to parameterization
        by applying `ParameterizedBaseHandler.filter` to all the handled
        callback queries. It also calls `pass_parameterized_handler`, which
        lets the callback functions get this handler as argument and deletes
        handled callback queries from the database relying on the passed
        identifiers.

        Parameters:
            client (Client): The client which is being set up.

        Warning:
            The functions the handlers make use of are not set up in the
            same way objects added using Pyrogram handlers are. Pyrubrum
            implements the following pattern::

                callback(handler, client, context, parameters)
        """
        for menu in self.get_menus():
            if menu.is_link:
                continue

            client.add_handler(
                CallbackQueryHandler(
                    pass_parameterized_handler(menu.on_callback, self),
                    self.filter(menu.menu_id),
                )
            )


def pass_parameterized_handler(
    func: Types.Callback, handler: ParameterizedBaseHandler
) -> Types.PyrogramCallback:
    """Generate a function which, whenever it is called, subsequently calls
    `callback`, passing the handler from which this object was generated and
    the retrieved parameters (it creates an empty dictionary if non existent).
    It requires a subclass of `ParameterizedBaseHandler` to be
    provided in order to work.

    Parameters:
        callback (Types.Callback): The callback function which
            automatically gets called by the generated function.
        handler (ParameterizedBaseHandler): The handler object which made
            use of this function in order to provide this one as a callback
            to a Pyrogram handler.

    Returns:
        Types.PyrogramCallback: The function which is being added to
        a Pyrogram handler.

    Warning:
        The functions the handlers make use of are not set up in the
        same way objects added using Pyrogram handlers are. Pyrubrum
        implements the following pattern::

            callback(handler, client, context, parameters)
    """

    def on_callback(client: Client, context):
        if isinstance(context, CallbackQuery):
            func(handler, client, context, context.parameters)
        else:
            func(handler, client, context, {})

    return on_callback



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from typing import Set

from pyrogram import Client
from pyrogram.handlers import MessageHandler

from pyrubrum.database import BaseDatabase
from pyrubrum.tree import Node

from .handler import Handler, command_filter, deep_link_filter
from .parameterized_base_handler import (
    ParameterizedBaseHandler,
    pass_parameterized_handler,
)


class ParameterizedHandler(Handler, ParameterizedBaseHandler):
    """Implementation of an handler which mixes the features of `Handler` and
    `ParameterizedBaseHandler` and has got, by definition, multiple
    :term:`top-level nodes <Top-level node>` whose linked menu are
    displayed to the user whenever a message is being handled and matches
    one of their filters, and a database with which it is possible to
    perform :term:`parameterization <Parameterization>`
    (i.e. it supports parameters).

    Parameters:
        nodes (Set[Node]): The :term:`top-level nodes <Top-level node>`,
            which represent the text commands that are available to the user.
            See `Handler` for more information.
        database (BaseDatabase): The storage for all the query parameters.
            It is used to pass parameters between menus. See
            `ParameterizedBaseHandler` for more information.
    """

    def __init__(self, nodes: Set[Node], database: BaseDatabase):
        Handler.__init__(self, nodes)
        ParameterizedBaseHandler.__init__(self, database)

    def setup(self, client: Client):
        """Make all the defined menus reachable by the client by adding handlers that
        catch all their identifiers to it. It adds support to parameterization
        by applying `ParameterizedBaseHandler.filter` to all the handled
        callback queries. It also calls `pass_parameterized_handler`, which
        lets the callback functions get this handler as argument and deletes
        handled callback queries from the database relying on the passed
        identifiers.

        Finally, it makes the :term:`top-level menus <Top-level menu>`
        reachable whenever a message is sent to the bot and matches one of
        their filters.

        Parameters:
            client (Client): The client which is being set up.

        Warning:
            The functions the handlers make use of are not set up in the
            same way objects added using Pyrogram handlers are. Pyrubrum
            implements the following pattern::

                callback(handler, client, context, parameters)
        """
        ParameterizedBaseHandler.setup(self, client)
        default_menu = None

        for node in self.nodes:
            if node.menu.is_link:
                continue

            if not node.menu.message_filter:
                node.menu.message_filter = command_filter(node.menu.menu_id)

            if node.menu.deep_link:
                node.menu.message_filter |= deep_link_filter(node.menu.menu_id)

            if node.menu.default:
                default_menu = node.menu

            client.add_handler(
                MessageHandler(
                    pass_parameterized_handler(node.menu.on_message, self),
                    node.menu.message_filter,
                )
            )

        if default_menu:
            client.add_handler(
                MessageHandler(
                    pass_parameterized_handler(default_menu.on_message, self)
                )
            )



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

import pytest

from pyrubrum import BaseMenu, Node, transform


def generate_menu(num: int) -> BaseMenu:
    return BaseMenu("Menu %d" % num, "menu_%d" % num)


menu_1 = generate_menu(1)
menu_2 = generate_menu(2)
menu_3 = generate_menu(3)
menu_4 = generate_menu(4)

final_node = Node(
    menu_1,
    children=(
        Node(menu_2, children=()),
        Node(menu_3, children=(Node(menu_4, children=()),)),
    ),
)

all_menus = {menu_1, menu_2, menu_3, menu_4}
children_menus = (menu_2, menu_3)

forest = transform({menu_1: {menu_2: None, menu_3: {menu_4}}})
nodes = iter(forest)
node = next(nodes)


def same_node(n1, n2):
    assert isinstance(n1.menu, BaseMenu)
    assert isinstance(n2.menu, BaseMenu)
    assert isinstance(n1.children, tuple)
    assert isinstance(n2.children, tuple)

    if hash(n1) != hash(n2):
        raise ValueError(
            "different hash for either n1 (%d) or n2 (%d)"
            % (hash(n1), hash(n2))
        )

    if hash(n1.menu) != hash(n2.menu):
        raise ValueError(
            "different hash for either n1.menu (%d) or "
            "n2.menu (%d)" % (hash(n1.menu), hash(n2.menu))
        )

    if len(n1.children) != len(n2.children):
        raise ValueError(
            "n1 and n2 do not share the same number of "
            "children (%d vs %d)" % (len(n1.children), hash(n2.children))
        )

    for i1, i2 in zip(n1.children, n2.children):
        same_node(i1, i2)


def test_transform():
    assert isinstance(forest, set)

    with pytest.raises(StopIteration):
        next(nodes)

    same_node(node, final_node)


def test_get_menus():
    got_menus = node.get_menus()

    assert isinstance(got_menus, set)
    assert got_menus == all_menus

    # do twice because to test cache
    got_menus = node.get_menus()

    assert isinstance(got_menus, set)
    assert got_menus == all_menus


def test_get_children_menus():
    got_menus = node.get_children_menus()

    assert isinstance(got_menus, tuple)
    assert got_menus == children_menus



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import List, Set

from pyrogram import Client
from pyrogram.filters import Filter, create
from pyrogram.handlers import CallbackQueryHandler
from pyrogram.types import InlineKeyboardButton

from pyrubrum.keyboard import Button
from pyrubrum.types import Types

NULL_POINTER = "0"


def callback_data_filter(match: str) -> Filter:
    return create(lambda _, __, m: m.data == match, "CallbackDataFilter")


class BaseHandler(ABC):
    """Basic representation of an handler, which is an entity that manages the
    setup of a `Client <pyrogram.Client>` instance (i.e. the bot).

    The purpose of this class is to give a general interface for an handler,
    even if it already implements both a sample setup and a keyboard processing
    function.

    See Also:
        For complete implementations of this class:

        * `Handler`
        * `ParameterizedHandler`

    Note:
        In order to create a subclass or to access this interface, you will
        need to implement the abstract method `get_menus`. Otherwise, you will
        get an error.
    """

    @abstractmethod
    def get_menus(self) -> Set["BaseMenu"]:
        """This abstract method is intended to be implemented as a collector
        of all the menus which were defined while initializing this object.

        Returns:
            Set[BaseMenu]: The set of the menus that were collected during
            the lookup process.
        """
        raise NotImplementedError

    def process_keyboard(
        self,
        keyboard: List[List[Button]],
        callback_query_id: str,
        chat_id: int,
    ) -> List[List[InlineKeyboardButton]]:
        """Given a list which represents an inline keyboard, a unique identifier
        for the callback and one for the chat from which the query was called,
        generate a Pyrogram-compatible inline keyboard.

        Parameters:
            keyboard (List[List[Button]]): The inline keyboard you want to
                process.
            callback_query_id (str): The unique identifier of the callback
                for which the keyboard is generated.
            chat_id (int): The identifier of the chat from which the query has
                been sent.

        Returns:
            List[List[pyrogram.InlineKeyboardButton]]: The generated keyboard
            in a Pyrogram-compatible type.
        """
        return [
            [
                (
                    InlineKeyboardButton(
                        button.name, callback_data=button.menu_id
                    )
                    if not button.link
                    else InlineKeyboardButton(button.name, url=button.link)
                )
                for button in row
            ]
            for row in keyboard
        ]

    def setup(self, client: Client):
        """Make all the defined menus reachable by the client by adding handlers that
        catch all their identifiers to it. It also calls `pass_handler`, which
        lets the callback functions get this handler as argument.

        Parameters:
            client (Client): The client which is being set up.

        Warning:
            The functions the handlers make use of are not set up in the
            same way objects added using Pyrogram handlers are. Pyrubrum
            implements the following pattern::

                callback(handler, client, context, parameters)
        """
        for menu in self.get_menus():
            if menu.is_link:
                continue

            client.add_handler(
                CallbackQueryHandler(
                    pass_handler(menu.on_callback, self),
                    callback_data_filter(menu.menu_id),
                )
            )


def pass_handler(
    callback: Types.Callback, handler: BaseHandler,
) -> Types.PyrogramCallback:
    """Generate a function which, whenever it is called, subsequently calls
    `callback`, passing the handler from which this object was generated.

    Parameters:
        callback (Types.Callback): The callback function which
            automatically gets called by the generated function.
        handler (BaseHandler): The handler object which made use of this
            function in order to provide this one as a callback to a
            Pyrogram handler.

    Returns:
        Types.PyrogramCallback: The function which is being added to
        a Pyrogram handler.

    Warning:
        The functions the handlers make use of are not set up in the
        same way objects added using Pyrogram handlers are. Pyrubrum
        implements the following pattern::

            callback(handler, client, context, parameters)
    """

    def on_callback(client: Client, context):
        callback(handler, client, context)

    return on_callback



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from functools import lru_cache
from typing import Iterable, Optional, Set, Tuple

from pyrogram import Client
from pyrogram.filters import Filter, create
from pyrogram.handlers import MessageHandler

from pyrubrum.menus import BaseMenu
from pyrubrum.tree import Node

from .base_handler import BaseHandler, pass_handler

DEEP_LINK_FILTER_TEMPLATE = "/start %s"


def command_filter(command: str) -> Filter:
    """Generate a filter that matches all the text messages following this
    pattern::

        /[COMMAND]

    Where ``COMMAND`` is the provided command. Unlike `Filters.command
    <pyrogram.Filters.command>`, it does not match any deep-linked or
    parameterized command.

    Parameters:
        command (str): The command that has to be matched, without
        the prefix ``/``.

    Returns:
        Filter: The generated filter.
    """
    return create(lambda _, __, m: m.text == "/" + command, "DeepLinkFilter")


def deep_link_filter(payload: str) -> Filter:
    """Generate a filter that matches all the text messages following this
    pattern::

        /start [PAYLOAD]

    Where ``PAYLOAD``is the provided payload.

    Parameters:
        payload (str): The payload that has to be matched.

    Returns:
        Filter: The generated filter.
    """
    match = DEEP_LINK_FILTER_TEMPLATE % payload
    return create(lambda _, __, m: m.text == match, "DeepLinkFilter")


class Handler(BaseHandler):
    """Implementation of a simple handler for non-parameterized menus which has
    got, by definition, multiple :term:`top-level nodes <Top-level node>`
    whose linked menus are displayed to the user whenever a message is being
    handled and matches one of their filters.

    Parameters:
        nodes (Set[Node]): The :term:`top-level nodes <Top-level node>`,
            which represent the text commands that are available to the user.

    Note:
        In order to make use of parameterized menus (e.g. `PageMenu`) or to
        pass parameters between menus, you will have to use an handler which
        supports such feature (e.g. `ParameterizedHandler`).
    """

    def __init__(self, nodes: Set[Node]):
        self.nodes = nodes

    @lru_cache()
    def get_family(
        self, menu_id: str
    ) -> Tuple[Optional[BaseMenu], Optional[Iterable[BaseMenu]]]:
        """Retrieve the menus which are linked to both parent and children of the
        :term:`top-level nodes <Top-level node>` of this instance if this
        instance matches the provided identifier. Otherwise it will search the
        menu matching it in the children of the
        :term:top-level nodes <Top-level node>` and will return their own
        families, if matched. On failure, it will return a tuple of length two
        filled with null values (i.e. ``None``).

        Parameters:
            menu_id (str): The identifier which must be matched.

        Returns:
            Tuple[Optional[BaseMenu], Optional[Iterable[BaseMenu]]]: A tuple of
            length two, whose first element is the parent node of the
            matched node while the second one is a tuple of all its children
            If no `Node` is found, the tuple will be filled with null
            values (i.e. ``None``).
        """
        for node in self.nodes:
            family = node.get_family(menu_id, None)

            if family[0] or family[1]:
                return family

        return (None, None)

    @lru_cache(maxsize=1)
    def get_menus(self) -> Set[BaseMenu]:
        """Retrieve the set of all the menus which are linked to the nodes belonging
        to the descent of the :term:`top-level nodes <Top-level node>` of this
        class (i.e. the children, the children of the children, etc...). In
        other words, it retrieves all the menus which were defined at the
        initialization of this instance.

        Returns:
            Set[BaseMenu]: The set of all the retrieved menus.
        """
        menus = set()

        for node in self.nodes:
            menus |= node.get_menus()

        return menus

    def setup(self, client: Client):
        """Set up a client instance by adding filters for handling callbacks and
        messages. If there is a default menu, it adds it using
        `MessageHandler <pyrogram.MessageHandler>`.

        Parameters:
            client (Client): The client you are setting up. It must be a bot
                instance in order to work.
        """
        BaseHandler.setup(self, client)
        default_menu = None

        for node in self.nodes:
            if node.menu.is_link:
                continue

            if not node.menu.message_filter:
                node.menu.message_filter = command_filter(node.menu.menu_id)

            if node.menu.deep_link:
                node.menu.message_filter |= deep_link_filter(node.menu.menu_id)

            if node.menu.default:
                default_menu = node.menu

            client.add_handler(
                MessageHandler(
                    pass_handler(node.menu.on_message, self),
                    node.menu.message_filter,
                )
            )

        if default_menu:
            client.add_handler(
                MessageHandler(pass_handler(default_menu.on_message, self))
            )



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Callable, Dict, Optional, Union

from pyrogram import Client
from pyrogram.filters import Filter
from pyrogram.types import (
    CallbackQuery,
    InlineKeyboardMarkup,
    InputMedia,
    Message,
)

from pyrubrum.keyboard import Keyboard
from pyrubrum.menus.styles import BaseStyle, MenuStyle
from pyrubrum.types import Types

from .base_menu import BaseMenu


class Menu(BaseMenu):
    """Implementation of a menu who has got, by definition, a content to display
    (i.e. what the user will see by accessing it), a name and a style.

    See Also:
        For complete examples of styles you can make use of:

        * `MenuStyle`
        * `PageStyle` (although it is recommended to use it only
          with `PageMenu`)

    Parameters:
        name (str): The name you give to the menu, which will be used as
            the text of callback button, if needed. See `BaseMenu` for more
            information.
        menu_id (str): The unique identifier given to the menu, which will
            refer unequivocally to this entity. The hash for this class is
            generated relying on the content of this field. Avoid using ``0``
            as it is used for buttons whose purpose is only related to
            design (i.e. they do not point to any menu,
            see :term:`Null-pointer button`). See `BaseMenu` for more
            information.
        content (Types.Content): What will be displayed whenever a user
            accesses this menu. Both text and media can be provided. A
            function that returns such values can be provided as well
            and must follow the this arguments pattern::

                func(handler, client, context, parameters)

        deep_link (Optional[bool]): If this menu shall be reached by a
            deep-link whose payload is the identifier of this instance.
            Defaults to ``False``.
        default (Optional[bool]): If this menu shall be displayed if no
            other :term:`top-level menu <Top-level menu>` has been matched.
            It works only if this menu is a :term:`top-level <Top-level menu>`
            one.
        message_filter (Optional[Filter]): A filter for telling Pyrogram
            when a message should be associated to this menu. It works only
            for :term:`top-level menus <Top-level menu>` (see `Handler.setup`).
            Defaults to ``None``, which automatically makes this menu reachable
            when the user texts a message that follows this pattern::

                /[MENU_ID]

        preliminary (Types.Preliminary): A function which is executed each time
            before doing anything else in `on_callback` and `on_message`.
            You can provide a list of such functions as well, which will be
            executed following the same order as the one of the list.
            Defaults to ``None``, which means that no function is going to
            be executed.
        style (BaseStyle): The class which generates the keyboard for this
            function following a certain style. Defaults to `MenuStyle()`.

    Note:
        This implementation is compatible with a non-parameterized handler.
    """

    def __init__(
        self,
        name: str,
        menu_id: str,
        content: Union[
            Union[InputMedia, str],
            Callable[
                [
                    "Handler",
                    Client,
                    Union[CallbackQuery, Message],
                    Dict[str, Any],
                ],
                Union[InputMedia, str],
            ],
        ],
        deep_link: Optional[bool] = False,
        default: Optional[bool] = False,
        message_filter: Optional[Filter] = None,
        preliminary: Types.Preliminary = None,
        style: BaseStyle = MenuStyle(),
        **kwargs
    ):
        BaseMenu.__init__(self, name, menu_id)
        self.content = content
        self.deep_link = deep_link
        self.default = default
        self.message_filter = message_filter
        self.preliminary = preliminary
        self.style = style

        for argument in kwargs:
            if not hasattr(self, argument):
                setattr(self, argument, kwargs[argument])

    def get_content(
        self,
        handler: "Handler",  # noqa
        client: Client,
        context: Union[CallbackQuery, Message],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Union[InputMedia, str]:
        """Get the content which will be displayed to the user.

        Parameters:
            handler (BaseHandler): The handler which coordinates the management
                of the menus.
            client (Client): The client which is linked to the handler.
            context (Union[CallbackQuery, Message]): The context for which the
                button is generated.
            parameters (Optional[Dict[str, Any]]): The parameters which were
                passed to the handler. Defaults to ``None``.

        Returns:
            Union[InputMedia, str]: The content of the menu, which is then
            displayed to the user as a media (if it is a subclass of
            `pyrogram.InputMedia`) or a message (if it is just a string).
        """
        if callable(self.content):
            if isinstance(parameters, dict):
                return self.content(handler, client, context, parameters)
            else:
                return self.content(handler, client, context)

        return self.content

    def on_callback(
        self,
        handler: "Handler",  # noqa
        client: Client,
        callback: CallbackQuery,
        parameters: Optional[Dict[str, Any]] = None,
    ):
        """Each time a callback query is handled, this function calls the preliminary
        function (i.e. `Menu.preliminary`), if callable, then gets the content
        that is going to be provided to the user (both text and media are
        compatible), sets up an inline keyboard using the style defined in
        `Menu.style` and finally edits the message with
        `CallbackQuery.edit_message_text` (if the content is a string, i.e. a
        text) or `CallbackQuery.edit_message_media`
        (if the content is an instance of `pyrogram.InputMedia`, i.e. a media).

        Parameters:
            handler (BaseHandler): The handler which coordinates the management
                of the menus.
            client (Client): The client which is linked to the handler.
            context (CallbackQuery): The callback query for which the button is
                generated.
            parameters (Optional[Dict[str, Any]]): The parameters which were
                passed to the handler. Defaults to ``None``.
        """

        if callable(self.preliminary):
            self.preliminary(handler, client, callback, parameters)
        elif isinstance(self.preliminary, list):
            for func in self.preliminary:
                func(handler, client, callback, parameters)

        content = self.get_content(handler, client, callback, parameters)

        if isinstance(content, InputMedia):
            callback.edit_message_media(
                content,
                reply_markup=self.keyboard(
                    handler, client, callback, parameters
                ),
            )
        elif isinstance(content, str):
            callback.edit_message_text(
                content,
                reply_markup=self.keyboard(
                    handler, client, callback, parameters
                ),
            )
        else:
            raise TypeError("content must be of type InputMedia or str")

    def on_message(
        self,
        handler: "Handler",  # noqa
        client: Client,
        message: Message,
        parameters: Optional[Dict[str, Any]] = None,
    ):
        """Each time a message is handled, this function calls the preliminary
        function (i.e. `Menu.preliminary`), if callable, then gets the content
        that is going to be provided to the user (both text and media are
        compatible), sets up an inline keyboard using the style defined in
        `Menu.style` and finally sends the message with `Message.reply_text`
        (if the content is a string, i.e. a text) or
        `Message.reply_cached_media` (if the content is an instance of
        `pyrogram.InputMedia`, i.e. a media).

        Parameters:
            handler (BaseHandler): The handler which coordinates the management
                of the menus.
            client (Client): The client which is linked to the handler.
            context (Message): The message for which the button is generated.
            parameters (Optional[Dict[str, Any]]): The parameters which were
                passed to the handler. Defaults to ``None``.
        """

        if callable(self.preliminary):
            self.preliminary(handler, client, message, parameters)
        elif isinstance(self.preliminary, list):
            for func in self.preliminary:
                func(handler, client, message, parameters)

        content = self.get_content(handler, client, message, parameters)

        if isinstance(content, InputMedia):
            message.reply_cached_media(
                file_id=content.media,
                file_ref=content.file_ref,
                caption=content.caption,
                reply_markup=self.keyboard(
                    handler, client, message, parameters
                ),
            )
        elif isinstance(content, str):
            message.reply_text(
                content,
                reply_markup=self.keyboard(
                    handler, client, message, parameters
                ),
            )
        else:
            raise TypeError("content must be of type InputMedia or str")

    def keyboard(
        self,
        handler: "Handler",  # noqa
        client: Client,
        context: Union[CallbackQuery, Message],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> InlineKeyboardMarkup:
        """Provide a keyboard using the chosen style in `Menu.style`.

        Parameters:
            handler (BaseHandler): The handler which coordinates the management
                of the menus.
            client (Client): The client which is linked to the handler.
            context (Union[CallbackQuery, Message]): The context for which the
                button is generated.
            parameters (Dict[str, Any]): The parameters which were passed to
                the handler.

        Returns:
            pyrogram.InlineKeyboardMarkup: The generated inline keyboard,
            which is then displayed to the user.
        """

        keyboard = self.style.generate(
            handler, client, context, parameters, self
        )

        if isinstance(context, Message):
            return (
                Keyboard(
                    keyboard,
                    handler,
                    str(context.message_id) + str(context.from_user.id),
                    context.chat.id,
                )
                if keyboard
                else None
            )
        elif isinstance(context, CallbackQuery):
            return (
                Keyboard(
                    keyboard, handler, context.id, context.message.chat.id
                )
                if keyboard
                else None
            )



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from environs import Env
from pyrogram import Client
from pyrogram.types import InputMediaPhoto

from pyrubrum import Handler, LinkMenu, Menu, transform

answer_image_link = "https://i.imgur.com/ZsDhqqt.png"
forty_two_image_link = "https://i.imgur.com/q0mZf3z.png"
wikipedia_link = (
    "https://en.wikipedia.org/wiki/Phrases_from"
    "_The_Hitchhiker%27s_Guide_to_the_Galaxy#The"
    "_Answer_to_the_Ultimate_Question_of_Life,_th"
    "e_Universe,_and_Everything_is_42"
)


def main(
    api_hash: str,
    api_id: int,
    bot_token: str,
    media_chat_id: int,
    session_name: str,
):
    bot = Client(
        session_name, api_hash=api_hash, api_id=api_id, bot_token=bot_token
    )

    with bot:
        answer_image_id = InputMediaPhoto(
            bot.send_photo(media_chat_id, answer_image_link).photo.file_id
        )
        forty_two_image_id = InputMediaPhoto(
            bot.send_photo(media_chat_id, forty_two_image_link).photo.file_id
        )

    tree = transform(
        {
            Menu("Start", "start", answer_image_id, default=True): {
                Menu("🌌 Get the answer", "forty_two", forty_two_image_id): {
                    LinkMenu("❓ Meaning", "meaning", wikipedia_link),
                },
            }
        }
    )

    handler = Handler(tree)
    handler.setup(bot)

    bot.run()


if __name__ == "__main__":
    env = Env()
    env.read_env()

    api_hash = env("API_HASH")
    api_id = env.int("API_ID")
    bot_token = env("BOT_TOKEN")
    media_chat_id = env.int("MEDIA_CHAT_ID")
    session_name = env("SESSION_NAME")

    main(api_hash, api_id, bot_token, media_chat_id, session_name)



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from abc import ABC
from typing import Any, Dict, Optional

from pyrogram import Client

from pyrubrum.handlers import BaseHandler
from pyrubrum.keyboard import Button


class BaseMenu(ABC):
    """Basic represention of a menu, which is an entity that has got by definition
    at least a name, a unique identifier and an associated button, which can
    be arbitrarily generated.

    A menu can represent a link as well (see `BaseMenu.is_link`), although it
    doesn't by default.

    The purpose of this class is to give a general interface for a menu, as it
    doesn't implement anything except for the generation of both buttons and
    hashes.

    See Also:
        For complete implementations of this class:

        * `LinkMenu`
        * `Menu`
        * `PageMenu`

    Parameters:
        name (str): The name you give to the menu, which will be used as the
            text of callback button, if needed.
        menu_id (str): The unique identifier given to the menu, which will
            refer unequivocally to this entity. The hash for this class is
            generated relying on the content of this field. Avoid using ``0``
            as it is used for buttons whose purpose is only related to design
            (i.e. they do not point to any menu, see :term:`Null-pointer
            button`).
        is_link (Optional[bool]): If this menu represents a link.

    Warning:
        Avoid using the same identifier for other entities, as it will result
        in ambiguity.
    """

    def __hash__(self) -> int:
        """The hash generator of a menu, relying on the unique identifier
        (`menu_id`) which is assigned to it.

        Returns:
            int: The unique hash which is generated for this entity.
        """
        return hash(self.menu_id)

    def __init__(
        self, name: str, menu_id: str, is_link: Optional[bool] = False
    ):
        self.is_link = is_link
        self.name = name
        self.menu_id = menu_id

    def button(
        self,
        handler: BaseHandler,
        client: Client,
        context: Any,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Button:
        """Create an inline button which refers to this menu, using `name` as
        the content of the text field and `menu_id` as the unique identifier
        of the `Button` object.

        Parameters:
            handler (BaseHandler): The handler which coordinates the management
                of the menus.
            client (Client): The client which is linked to the handler.
            context (Union[CallbackQuery, Message]): The context for which
                button is generated for.
            parameters (Optional[Dict[str, Any]]): The parameters which were
                passed to the handler. Defaults to ``None``.

        Returns:
            Button: The generated button.
        """
        return Button(self.name, self.menu_id, parameters)



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Dict, Optional

from pyrogram import Client

from pyrubrum.keyboard.button import Button
from pyrubrum.types import Types

from .base_menu import BaseMenu

DEEP_LINK_TEMPLATE = "https://t.me/{username}?{deep_link_type}={payload}"
PERMITTED_TYPES = {"start", "startgroup"}
USERNAMES = {}


class DeepLinkMenu(BaseMenu):
    """Implementation of a menu which automatically creates a deep-link,
    given a parameter that shall be passed.

    See also:
        `What the Bot API documentation says about deep-linking.
        <https://core.telegram.org/bots#deep-linking>`_

    Parameters:
        name (str): The name you give to the menu, which will be used as
            the text of callback button, if needed. See `BaseMenu` for more
            information.
        menu_id (str): The unique identifier given to the menu, which will
            refer unequivocally to this entity. The hash for this class is
            generated relying on the content of this field. Avoid using ``0``
            as it is used for buttons whose purpose is only related to
            design (i.e. they do not point to any menu, see
            :term:`Null-pointer button`). See `BaseMenu` for more information.
        payload (Types.Payload): What will be passed to as payload using the
            deep-linking. A function can be provided as well and must follow
            the following arguments pattern::

                func(handler, client, context, parameters)
        deep_link_type (Optional[str]): The type of deep-link that is being
            generated, which must be either ``start`` or ``startgroup``.
            Defaults to ``start``.
    """

    def __init__(
        self,
        name: str,
        menu_id: str,
        payload: Types.Payload,
        deep_link_type: Optional[str] = "start",
    ):
        if deep_link_type not in PERMITTED_TYPES:
            raise ValueError(
                "deep_link_type must be either " "'start' or 'startgroup'"
            )

        BaseMenu.__init__(self, name, menu_id, is_link=True)

        self.payload = payload
        self.deep_link_type = deep_link_type

    def button(
        self,
        handler: "BaseHandler",  # noqa
        client: "Client",  # noqa
        context: Any,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Button:
        """Create an inline button which makes use of deep-linking, following
        the defined `DeepLinkMenu.deep_link_type` and passing the parameter
        retrieved from `DeepLinkMenu.payload`.

        Parameters:
            handler (BaseHandler): The handler which coordinates the management
                of the menus.
            client (Client): The client which is linked to the handler.
            context (Union[CallbackQuery, Message]): The context for which
                button is generated for.
            parameters (Optional[Dict[str, Any]]): The parameters which were
                passed to the handler. Defaults to ``None``.

        Returns:
            Button: The generated button.
        """
        client_hash = hash(client)

        if client_hash in USERNAMES:
            username = USERNAMES[client_hash]
        else:
            username = client.get_me().username
            USERNAMES[client_hash] = username

        if callable(self.payload):
            return Button(
                self.name,
                self.menu_id,
                link=DEEP_LINK_TEMPLATE.format(
                    deep_link_type=self.deep_link_type,
                    payload=self.payload(handler, client, context, parameters),
                    username=username,
                ),
            )

        return Button(
            self.name,
            self.menu_id,
            link=DEEP_LINK_TEMPLATE.format(
                deep_link_type=self.deep_link_type,
                payload=self.payload,
                username=username,
            ),
        )



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from .error import DatabaseError


class GetError(DatabaseError):
    """Exception which is raised whenever an error occurs while getting a certain
    key from the database.
    """

    def __init__(self):
        super().__init__(
            "An error occured while getting a key from the database"
        )



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from typing import List

from pyrogram.types import InlineKeyboardMarkup

from .button import Button


class Keyboard(InlineKeyboardMarkup):
    """Simple shortcut for calling `BaseHandler.process_keyboard` and storing the
    processed keyboard in the same object.

    It is useful since it automatically processes the content of buttons
    and, at the same time, behaves like a normal `InlineKeyboardMarkup
    <pyrogram.InlineKeyboardMarkup>` instance.

    Parameters:
        inline_keyboard (List[List[Button]]): The keyboard which is going
            to be converted.
        handler (BaseHandler): The handler from which the function
            `BaseHandler.process_function` is going to be called.
        callback_query_id (str): The unique identifier of the query, which
            will be used to make the call that is made by the inline
            buttons unique.
        chat_id (int): The identifier of the chat from which the query has
            been sent.
    """

    def __init__(
        self,
        inline_keyboard: List[List[Button]],
        handler: "BaseHandler",  # noqa
        callback_query_id: str,
        chat_id: int,
    ):
        super().__init__(
            handler.process_keyboard(
                inline_keyboard, callback_query_id, chat_id
            )
        )



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from typing import Optional

from pyrubrum.types import Types


class BaseDatabase(ABC):
    """Basic representation of a database, which, by definition, implements the
    three fundamental operations `delete`, `get` and `set`.

    The purpose of this class is to give a general interface for a database,
    as it does not implement anything.

    A sample implementation of this interface is `~pyrubrum.RedisDatabase`.

    Note:
        In order to create a subclass or to access this interface, you will
        need to implement all the abstract methods, which are `delete`, `get`
        and `set`. Otherwise, you will get an error.
    """

    @abstractmethod
    def get(self, key: str) -> str:
        """This abstract method is intended to be implemented in order to get the value
        which is stored with a certain key inside the database.

        Parameters:
            key (str): The key you are retrieving the value of.

        Returns:
            Optional[str]: The value which is associated to the key.

        Raises:
            GetError: If an error occured while retrieving the key from the
                database.
            NotFoundError: If the provided key is not found.
        """
        raise NotImplementedError

    @abstractmethod
    def set(self, key: str, value: str, expire: Optional[Types.Expire] = None):
        """This abstract method is intended to be implemented in order to assign a
        value to a certain key inside the database. It may even be marked with
        an expire as to avoid having too much unused data stored inside the
        database.

        Parameters:
            key (str): The key you are adding or updating the value of.
            value (str): The value which is being assigned to the key.
            expire (Optional[Types.Expire]): The expire in seconds or as a
                `timedelta` object. A key is set not to expire if ``False`` is
                provided for this argument. Defaults to ``None``.

        Raises:
            ExpireError: If an error occured while setting the expire for the
                key.
            SetError: If an error occured while inserting the key into the
                database.
        """
        raise NotImplementedError

    @abstractmethod
    def delete(self, key: str):
        """This abstract method is intended to be implemented in order to delete a
        certain key from the database, together with its stored value.

        Parameters:
            key (str): The key which is being deleted from the database,
                together with its linked data.

        Raises:
            DeleteError: If an error occured while deleting the key.
            NotFoundError: If the provided key is not found.
        """
        raise NotImplementedError



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict

from pyrogram import Client

from pyrubrum.database.base_database import BaseDatabase
from pyrubrum.database.errors.not_found_error import NotFoundError

try:
    import orjson as json  # noqa
except (ImportError, ModuleNotFoundError):
    import json


class User(dict):
    """Representation of a user, which behaves like a dictionary and
    has got, by definition, a key named `user_id` which is the unique
    identifier of the user.
    """

    def delete(self, database: BaseDatabase):
        """Given a database, delete this object from it.

        Parameters:
            database (BaseDatabase): The database from which the user is being
                deleted.
        """
        database.delete("user_" + str(self["user_id"]))

    @classmethod
    def get(cls, database: BaseDatabase, user_id: int) -> "User":
        """Get a user from a given database seeking a key that follows
        this pattern::

            user_[USER_ID]

        Parameters:
            database (BaseDatabase): The database which is queried to seek
                the user.
            user_id (int): The unique identifier of the user that is being
                seeked.

        Returns:
            User: The retrieved user.
        """
        return cls.parse(database.get("user_" + str(user_id)))

    @classmethod
    def get_or_create(cls, database: BaseDatabase, user_id: int) -> "User":
        """Get a user from a given database, if any. Otherwise, a new user
        is created using the following pattern as key::

            user_[USER_ID]

        Parameters:
            database (BaseDatabase): The database which is queried to seek
                the user.
            user_id (int): The unique identifier of the user that is being
                seeked.

        Returns:
            User: The retrieved user.
        """
        try:
            return cls.parse(database.get("user_" + str(user_id)))
        except NotFoundError:
            user = cls({"user_id": user_id})
            user.save(database)
            return user

    @classmethod
    def parse(cls, data: str) -> "User":
        """Automatically create a new `User` object from a JSON object.

        Parameters:
            data (str): The JSON object as string.

        Returns:
            User: The generated `User` object.
        """

        return cls(json.loads(data))

    @classmethod
    def preliminary(
        cls,
        handler: "ParameterizedBaseHandler",  # noqa
        client: Client,
        context: Any,
        parameters: Dict[str, Any] = None,
    ):
        """Automatically retrieve the `User` object relying on the identifier of the
        user who made the query and assign it to ``context.current_user``.

        Parameters:
            handler (BaseHandler): The handler which coordinates the management
                of the menus.
            client (Client): The client which is linked to the handler.
            context (CallbackQuery): The callback query for which the button is
                generated.
            parameters (Optional[Dict[str, Any]]): The parameters which were
                passed to the handler. Defaults to ``None``.
        """
        context.current_user = cls.get_or_create(
            handler.database, context.from_user.id
        )

    def save(self, database: BaseDatabase):
        """Save the `User` object to the the database using the following pattern
        as key::

            user_[USER_ID]

        Parameters:
            database (BaseDatabase): The database to which this object is
                being saved.
        """
        database.set("user_" + str(self["user_id"]), json.dumps(self))



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Dict, Optional

from pyrogram import Client

from pyrubrum.keyboard.button import Button
from pyrubrum.types import Types

from .base_menu import BaseMenu


class LinkMenu(BaseMenu):
    """Implementation of a menu which represents a link to a website.

    Parameters:
        name (str): The name you give to the menu, which will be used as
            the text of callback button, if needed. See `BaseMenu` for more
            information.
        menu_id (str): The unique identifier given to the menu, which will
            refer unequivocally to this entity. The hash for this class is
            generated relying on the content of this field. Avoid using ``0``
            as it is used for buttons whose purpose is only related to
            design (i.e. they do not point to any menu, see
            :term:`Null-pointer button`). See `BaseMenu` for more information.
        link (Types.Content): What will be displayed whenever a user
            accesses this menu. A function can be provided as well and must
            follow the following arguments pattern::

                func(handler, client, context, parameters)

    """

    def __init__(self, name: str, menu_id: str, link: Types.Link):
        BaseMenu.__init__(self, name, menu_id, is_link=True)

        self.link = link

    def button(
        self,
        handler: "BaseHandler",  # noqa
        client: "Client",  # noqa
        context: Any,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Button:
        """Create an inline button which redirects to a website retrieved from
        `LinkMenu.link`, using `name` as the content of the text field.

        Parameters:
            handler (BaseHandler): The handler which coordinates the management
                of the menus.
            client (Client): The client which is linked to the handler.
            context (Union[CallbackQuery, Message]): The context for which
                button is generated for.
            parameters (Optional[Dict[str, Any]]): The parameters which were
                passed to the handler. Defaults to ``None``.

        Returns:
            Button: The generated button.
        """
        if callable(self.link):
            return Button(
                self.name,
                self.menu_id,
                link=self.link(handler, client, context, parameters),
            )

        return Button(self.name, self.menu_id, link=self.link)



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

__author__ = "Hearot"
__author_email__ = "gabriel@hearot.it"
__copyright__ = "Copyright (C) 2020 Hearot <https://github.com/hearot>"
__license__ = "GNU General Public License v3"
__package__ = "pyrubrum"
__url__ = "https://github.com/hearot/pyrubrum"
__version__ = "0.1a3"

from .database import BaseDatabase  # noqa
from .database import DictDatabase  # noqa
from .database import RedisDatabase  # noqa
from .database.errors import DatabaseError  # noqa
from .database.errors import DeleteError  # noqa
from .database.errors import ExpireError  # noqa
from .database.errors import GetError  # noqa
from .database.errors import NotFoundError  # noqa
from .database.errors import SetError  # noqa
from .handlers import BaseHandler  # noqa
from .handlers import Handler  # noqa
from .handlers import ParameterizedBaseHandler  # noqa
from .handlers import ParameterizedHandler  # noqa
from .handlers import pass_handler  # noqa
from .handlers import pass_parameterized_handler  # noqa
from .keyboard import Button  # noqa
from .keyboard import Element  # noqa
from .keyboard import Keyboard  # noqa
from .menus import BaseMenu  # noqa
from .menus import DeepLinkMenu  # noqa
from .menus import LinkMenu  # noqa
from .menus import Menu  # noqa
from .menus import PageMenu  # noqa
from .menus.styles import BaseStyle  # noqa
from .menus.styles import MenuStyle  # noqa
from .menus.styles import PageStyle  # noqa
from .objects import User  # noqa
from .tree import Node  # noqa
from .tree import recursive_add  # noqa
from .tree import transform  # noqa
from .types import Types  # noqa



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from pyrubrum import Menu


class FakeContext:
    @staticmethod
    def edit_message_media(*a, **b):
        pass

    @staticmethod
    def edit_message_text(*a, **b):
        pass

    @staticmethod
    def reply_cached_media(*a, **b):
        pass

    @staticmethod
    def reply_text(*a, **b):
        pass


class FakeHandler:
    @staticmethod
    def get_family(*a, **b):
        return (None, None)


context = FakeContext()
handler = FakeHandler()


def test_menu():
    menu = Menu("Test", "test", "test_content")
    menu.on_callback(handler, None, context)
    menu.on_message(handler, None, context)


def test_menu_with_function():
    worked = []

    def get_content(a, b, c, d=None):
        worked.append(1)
        return "test_content"

    menu = Menu("Test", "test", get_content)
    menu.on_callback(handler, None, context)
    menu.on_message(handler, None, context)

    assert len(worked) == 2


def test_preliminary():
    worked = []

    def preliminary(a, b, c, d=None):
        worked.append(1)

    menu = Menu("Test", "test", "test_content", preliminary=preliminary)
    menu.on_callback(handler, None, context)
    menu.on_message(handler, None, context)

    menu = Menu("Test", "test", "test_content", preliminary=[preliminary] * 5)
    menu.on_callback(handler, None, context)
    menu.on_message(handler, None, context)

    menu = Menu("Test", "test", "test_content", preliminary=None)
    menu.on_callback(handler, None, context)
    menu.on_message(handler, None, context)

    assert len(worked) == 12



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from datetime import timedelta
from typing import Any, Callable, Dict, List, Optional, Union

from pyrogram import Client
from pyrogram.types import InputMedia

from pyrubrum.keyboard.element import Element


class Types:
    Callback = Callable[
        ["BaseHandler", Client, Any, Optional[Dict[str, Any]]], None
    ]
    """This type defines the possible values for functions that work as
    callbacks in Pyrubrum (e.g. `on_callback`, `on_message`). They must
    implement the following pattern to be valid::

        callback(handler, client, context, parameters=None)
    """

    Content = Union[
        Union[InputMedia, str],
        Callable[
            ["Handler", Client, Any, Dict[str, Any]], Union[InputMedia, str],
        ],
    ]
    """This type defines the possible values for the content that is provided as
    argument to a menu. A string or an instance of `InputMedia
    <pyrogram.InputMedia>` is suitable for this type to be valid.
    A function that returns such values is valid as well and must follow this
    pattern::

        content(handler, client, context, parameters=None)
    """

    Expire = Optional[Union[bool, int, timedelta]]
    """This type defines the possible values for an expire (see `BaseDatabase`).
    A positive integer is suitable and indicates how many seconds a value can
    be kept within database. A `datetime.timedelta` can be provided and is
    valid as well. In order to indicate that there is no expire, ``False``
    shall be provided.
    """

    Items = Union[
        List[Element],
        Callable[
            ["ParameterizedHandler", Client, Any, Dict[str, Any]],
            List[Element],
        ],
    ]
    """This type defines the possible values for items that are provided as
    argument to `PageMenu`. A list of `Element` is suitable for this type.
    A function that returns such list is valid as well and must follow this
    pattern::


        items(handler, client, context, parameters)
    """

    Link = Union[
        str, Callable[["Handler", Client, Any, Dict[str, Any]], str],
    ]
    """This type defines the possible values for the link that is used to build
    an inline button that redirects to a website. A string is suitable for this
    type to be valid.
    A function that returns such value is valid as well and must follow this
    pattern::

        link(handler, client, context, parameters=None)
    """

    Payload = Union[
        str, Callable[["Handler", Client, Any, Dict[str, Any]], str],
    ]
    """This type defines the possible values for the parameter that is passed
    to the bot using a deep-link. A string is suitable for this type to be
    valid.
    A function that returns such value is valid as well and must follow this
    pattern::

        payload(handler, client, context, parameters=None)
    """

    Preliminary = Optional[
        Union[
            Callable[["Handler", Client, Any, Dict[str, Any]], None],
            List[Callable[["Handler", Client, Any, Dict[str, Any]], None]],
        ]
    ]
    """This type defines the possible values which can be passed for a
    preliminary function. All the provided functions for this type must
    follow this pattern::

        preliminary(handler, client, context, parameters=None)

    A list of such functions can be provided as well for this type to
    be valid.
    """

    PyrogramCallback = Callable[[Client, Any], None]
    """This type describes the pattern of functions which are added to
    a Pyrogram handler, being it::

        callback(client, context)
    """



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from itertools import islice
from typing import Any, Dict, List, Optional

from pyrogram import Client

from .base_style import BaseStyle


class MenuStyle(BaseStyle):
    """Provide a standard keyboard with which you can refer to children and
    parent menus effortlessly.

    Parameters:
        back_text (Optional[str]): The text which will be displayed
            inside the button that lets the user go back to the parent
            menu. Defaults to "🔙".
        limit (Optional[int]): The limit of buttons per row. Defaults to ``2``.
    """

    def __init__(
        self, back_text: Optional[str] = "🔙", limit: Optional[int] = 2,
    ):
        self.back_text = back_text
        self.limit = limit

    def generate(
        self,
        handler: "Handler",  # noqa
        client: Client,
        context: Any,
        parameters: Dict[str, Any],
        menu: "Menu",  # noqa
    ) -> List[List["Button"]]:
        """Provide a keyboard, filled with all the buttons which refer to the
        children menus and a special button for linking the user to the parent
        menu, if any.

        Parameters:
            handler (BaseHandler): The handler which coordinates the management
                of the menus.
            client (Client): The client which is linked to the handler.
            context (Union[CallbackQuery, Message]): The context for which the
                button is generated.
            parameters (Dict[str, Any]): The parameters which were passed to
                the handler.
            menu (Menu): The menu the keyboard is being built for.

        Returns:
            pyrogram.InlineKeyboardMarkup: The generated inline keyboard,
            which is then displayed to the user.
        """
        parent, children = handler.get_family(menu.menu_id)

        keyboard = []

        if children:
            iterable = iter(children)
            keyboard = list(
                iter(
                    lambda: list(
                        map(
                            lambda child: child.button(
                                handler, client, context, parameters
                            ),
                            islice(iterable, self.limit),
                        )
                    ),
                    [],
                )
            )

        if parent:
            parent_button = parent.button(handler, client, context, parameters)
            parent_button.name = self.back_text

            keyboard = keyboard + [[parent_button]]

        return keyboard



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from functools import lru_cache
from typing import Any, Dict, Iterable, Optional, Set, Tuple, Union

from pyrubrum.menus import BaseMenu

Family = Tuple[Optional[BaseMenu], Optional[Set[BaseMenu]]]
MenuIterable = Union[Dict[BaseMenu, Any], Iterable[BaseMenu]]


class Node:
    """Representation of a single object in a tree, which might have defined a
    parent (i.e. the `Node` object it is child of) and a set of other nodes,
    which are its children. Each `Node` instance is linked to a subclass of
    `BaseMenu`.

    Parameters:
        menu (BaseMenu): The menu this node is being linked to.
        children (Optional[Set[Node]]): The nodes whose parent is going
            to be this instance. Defaults to ``None``, which makes the
            children set an empty one.
    """

    def __hash__(self) -> int:
        """The hash generator for a node, relying on the hash of the linked
        menu.

        Returns:
            int: The hash of the menu which is linked to this instance.
        """
        return hash(self.menu)

    def __init__(self, menu: BaseMenu, children: Optional[Set["Node"]] = None):
        self.children = children if children else tuple()
        self.menu = menu

    def add_child(self, node: "Node"):
        """Add a `Node` instance to the set of children.

        Parameters:
            node (Node): The node which is being added as a child of this
                object.
        """
        self.children += (node,)

    @lru_cache(maxsize=1)
    def get_children_menus(self) -> Iterable[BaseMenu]:
        """Get all the menus that are linked to the children belonging to
        this instance.

        Returns:
            Iterable[BaseMenu]: A tuple that contains all the retrieved menus.
        """
        children = tuple(child.menu for child in self.children)
        return children if children else None

    @lru_cache()
    def get_family(self, menu_id: str, parent: Optional["Node"]) -> Family:
        """Retrieve the menus which are linked to both parent and children of this
        instance if this instance matches the provided identifier. Otherwise it
        will search the menu matching it in its children and return its family,
        if matched. On failure, it will return a tuple of length two filled
        with null values (i.e. ``None``).

        Parameters:
            menu_id (str): The identifier which must be matched.
            parent (Optional[Node]): The parent this ``Node`` comes from.

        Returns:
            Tuple[Optional[BaseMenu], Optional[Tuple[BaseMenu]]]: A tuple of
            length two, whose first element is the parent node of the
            matched node while the second one is a tuple of all its children
            If no `Node` is found, the tuple will be filled with null
            values (i.e. ``None``).
        """
        if self.menu.menu_id == menu_id:
            return (
                parent.menu if isinstance(parent, Node) else None,
                self.get_children_menus(),
            )

        for child in self.children:
            child_menus = child.get_family(menu_id, self)

            if child_menus[0]:
                return child_menus

        return (None, None)

    @lru_cache(maxsize=1)
    def get_menus(self) -> Set[BaseMenu]:
        """Retrieve the set of all the menus which are linked to the nodes belonging
        to the descent of this class (i.e. the children, the children of the
        children, etc...).

        Returns:
            Set[BaseMenu]: The set of all the retrieved menus.
        """
        menus = {self.menu}

        for child in self.children:
            if child.menu.menu_id != self.menu.menu_id:
                menus |= child.get_menus()

        return menus


def recursive_add(menus: MenuIterable, parent: Node):
    """Link the provided menus to `Node` instances and add these ones to a provided
    parent. Finally, for each found value which is iterable, this function is
    called in a recursive way in order to link its elements to its parent
    `Node`.

    Parameters:
        menus (MenuIterable): The provided iterable of the menus which are
            being added to the parent node. If it is a dictionary, the stored
            value for each key, if iterable, will be recursively provided as
            argument to a new call of this function, together with the parent
            node, which is the key.
        parent (Node): The parent the nodes are being added to.
    """
    for menu in menus:
        node = Node(menu)
        parent.add_child(node)

        if isinstance(menus, dict) and isinstance(menus[menu], Iterable):
            recursive_add(menus[menu], node)


def transform(menus: MenuIterable) -> Set[Node]:
    """Transform an iterable compounded of menus into a `Node` object. If a
    dictionary is provided as argument, its values will be transformed as
    well and added as children using ``recursive_add``.

    Parameters:
        menus (Union[Dict[BaseMenu, Any], Iterable[BaseMenu]]): The iterable
            whose elements are `BaseMenu` instances.

    Returns:
        Node: The `Node` object which collects all the menus as children.
    """
    nodes = set()

    for menu in menus:
        node = Node(menu)
        nodes.add(node)

        if isinstance(menus, dict) and menus[menu]:
            recursive_add(menus[menu], node)

    return nodes



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from copy import deepcopy
from typing import Any, Dict, Optional


class Button:
    """Representation of a button which belongs to an inline keyboard and has got,
    by definition, a unique identifier, which actually represents the menu this
    object is referring to, an `Element` identifier (i.e. the identifier of the
    object that is carried by the button and is then passed to the menu it is
    referring to, if any), a name, which is displayed inside the text field of
    the inline button, and a dictionary of parameters.

    The dictionary that is passed as argument for representing parameters is
    always deep copied, if not ``None``.

    Parameters:
        name (str): The name which is displayed inside the text field of
            the inline button this object will be converted to.
        menu_id (str): The unique identifier of the button, which is the
            same as the one of the menu the button is referring to.
        parameters (Optional[Dict[str, Any]]): The parameters that will be
            passed to the menu this button is referring to. Defaults to
            ``None`` (i.e. the handler does not support parameterization).
        element_id (Optional[str]): The unique identifier of the `Element`
            the button is carrying, if any. Defaults to an empty string.
        link (Optional[str]): The website this button is referring to, if any.
            If provided, all the other arguments are ignored and this button
            is set to represent a button which redirects to a certain URL,
            which is the provided string itself.
        same_menu (Optional[bool]): If the button is referring to the same
            menu by which it was initialized. Defaults to ``False``.
        **kwargs: Arbitrary keyword arguments which can be used as a way to
            define a new parameter (``key=value``) that will be added to
            the dictionary of parameters.

    Warning:
        There is a special group of names which are not available for being
        used as keys for parameters (not for arguments!). These are
        ``from_chat_id``, ``menu_id``, ``element_id`` and ``same_menu``.
        If provided, they will be overwritten. Keys starting with ``page_``
        should be avoided according to `PageStyle.generate_page_id`.
    """

    def __init__(
        self,
        name: str,
        menu_id: str,
        parameters: Optional[Dict[str, Any]] = None,
        element_id: Optional[str] = "",
        link: Optional[str] = None,
        same_menu: Optional[bool] = False,
        **kwargs
    ):
        self.name = name
        self.link = link

        if link:
            return

        self.menu_id = menu_id
        self.element_id = element_id
        self.same_menu = same_menu

        if isinstance(parameters, dict):
            self.parameters = deepcopy(parameters)
            self.parameters.update(kwargs)
        else:
            self.parameters = None



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pyrubrum. If not, see <http://www.gnu.org/licenses/>.

from pyrubrum import LinkMenu

example_link = "https://example.com"
example_name = "Example"


def test_button():
    link = LinkMenu(example_name, "test", example_link)
    button = link.button(None, None, None)

    assert button.link == example_link
    assert not hasattr(button, "menu_id")
    assert button.name == example_name


def test_button_with_function():
    link = LinkMenu(example_name, "test", lambda a, b, c, d: example_link)
    button = link.button(None, None, None)

    assert button.link == example_link
    assert not hasattr(button, "menu_id")
    assert button.name == example_name



#!/usr/bin/env python3
# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

import re

from setuptools import find_packages
from setuptools import setup

GITHUB_REPOSITORY = "https://github.com/hearot/pyrubrum/blob/v%s/"


with open("fast-requirements.txt", encoding="utf-8") as r:
    fast_requirements = [p.strip() for p in r]

with open("pyrubrum/__init__.py", encoding="utf-8") as f:
    version = re.findall(r"__version__ = \"(.+)\"", f.read())[0]

with open("README.md", encoding="utf-8") as f:
    long_description = f.read().replace("./", GITHUB_REPOSITORY % version)

with open("requirements.txt", encoding="utf-8") as r:
    requirements = [p.strip() for p in r]

setup(
    author="Hearot",
    author_email="gabriel@hearot.it",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Communications",
        "Topic :: Communications :: Chat",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Typing :: Typed",
    ],
    description="An intuitive framework for creating Telegram bots",
    extras_require={"fast": fast_requirements},
    install_requires=requirements,
    keywords="bot bots chat messenger mtproto pyrogram python telegram",
    license="GPLv3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    name="Pyrubrum",
    packages=find_packages(),
    project_urls={
        "Tracker": "https://github.com/hearot/pyrubrum/issues",
        "Source": "https://github.com/hearot/pyrubrum",
    },
    python_requires=">=3.6.*",
    url="https://github.com/hearot/pyrubrum",
    version=version,
)



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from calendar import monthrange
from datetime import datetime
from typing import Set, Union

from environs import Env
from pyrogram import Client

from pyrubrum import (
    DictDatabase,
    Element,
    Menu,
    Node,
    PageMenu,
    PageStyle,
    ParameterizedHandler,
    RedisDatabase,
    transform,
)

try:
    from redis import Redis
except (ImportError, ModuleNotFoundError):
    pass


def generate_days(handler, client, context, parameters):
    month = int(parameters["month_id"])
    year = int(parameters["year_id"])

    return [
        Element(str(day + 1), str(day + 1))
        for day in range(monthrange(year, month)[1])
    ]


def generate_months():
    return [Element(str(month + 1), str(month + 1)) for month in range(12)]


def generate_years(start: int, end: int):
    return [Element(str(year), str(year)) for year in range(start, end + 1)]


def tell_about_the_day(handler, client, context, parameters):
    day = int(parameters["day_id"])
    month = int(parameters["month_id"])
    year = int(parameters["year_id"])

    return "📅 " + datetime(year, month, day).strftime("%d/%m/%Y is a %A.")


tree = transform(
    {
        PageMenu(
            "Start",
            "start",
            "📅 Choose a year.",
            generate_years(1970, 2044),
            default=True,
            style=PageStyle(limit=4, limit_items=15),
        ): {
            PageMenu(
                "After year choice",
                "year",
                "📅 Choose a month.",
                generate_months(),
                style=PageStyle(limit=5, limit_items=12),
            ): {
                PageMenu(
                    "After month choice",
                    "month",
                    "📅 Choose a day.",
                    generate_days,
                    style=PageStyle(limit=5, limit_items=31),
                ): {Menu("After day choice", "day", tell_about_the_day)}
            }
        }
    }
)


def main(
    api_hash: str,
    api_id: int,
    bot_token: str,
    database: Union[DictDatabase, RedisDatabase],
    session_name: str,
    tree: Set[Node],
):
    bot = Client(
        session_name, api_hash=api_hash, api_id=api_id, bot_token=bot_token
    )

    handler = ParameterizedHandler(tree, database)
    handler.setup(bot)

    bot.run()


if __name__ == "__main__":
    env = Env()
    env.read_env()

    api_hash = env("API_HASH")
    api_id = env.int("API_ID")
    bot_token = env("BOT_TOKEN")
    session_name = env("SESSION_NAME")

    if env.bool("USE_REDIS", False):
        db = env.int("REDIS_DB")
        host = env("REDIS_HOST")
        port = env.int("REDIS_PORT")
        database = RedisDatabase(Redis(host=host, port=port, db=db))
    else:
        database = DictDatabase()

    main(api_hash, api_id, bot_token, database, session_name, tree)



# Pyrubrum - An intuitive framework for creating Telegram bots
# Copyright (C) 2020 Hearot <https://github.com/hearot>
#
# This file is part of Pyrubrum.
#
# Pyrubrum is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyrubrum is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyrubrum. If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from pyrubrum.types import Types

from .base_database import BaseDatabase
from .errors import NotFoundError


class DictDatabase(dict, BaseDatabase):
    """Reduced implementation of a database using a dictionary, without the
    assignment of an expire to a key whenever one is added to the database.

    Warning:
        It not recommended to use this in production, as it does not implement
        expires, while other implementations, such as `RedisDatabase
        <pyrubrum.RedisDatabase>`, do. In addition, any stored data will be
        erased as soon as the program stops executing. This implementation
        might be useful only in development and testing mode.
    """

    def get(self, key: str) -> str:
        """Get the value which is associated to a certain key inside the database.

        This method will query the key using `dict.get`. If the key is not
        defined within the dictionary, `NotFoundError` is raised.

        Parameters:
            key (str): The key you are retrieving the value of from the
                dictionary.

        Returns:
            Optional[str]: The value which is associated to the key in the
            dictionary.

        Raises:
            NotFoundError: If the provided key is not found.
        """
        if key not in self:
            raise NotFoundError(key)

        return dict.get(self, key)

    def set(self, key: str, value: str, expire: Optional[Types.Expire] = None):
        """Assign a value to a certain key inside the database. Note that this
        implementation ignores the setting of any expire.

        This method will assign the provided value to the key using
        `dict.update`.

        Parameters:
            key (str): The key you are adding or updating the value of.
            value (str): The value which is being assigned to the key.
            expire (Optional[Types.Expire]): It gets ignored by this
                implementation. Defaults to ``None``.
        """

        self.update({key: value})

    def delete(self, key: str):
        """Delete a certain key from the database, together with its stored value.

        This method will delete the provided key from the database using
        `dict.pop`.

        Parameters:
            key (str): The key which is being deleted from the database,
                together with its linked data.

        Raises:
            NotFoundError: If the provided key is not found.
        """
        try:
            self.pop(key)
        except KeyError:
            raise NotFoundError(key)
